/* ia32-trace.c - ia32 control flow tracer
 *
 * by scut
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <common.h>
#include <ia32-decode.h>
#include <ia32-trace.h>
#include <ia32-function.h>


/*** static prototypes
 */

static ia32_branch **
ia32_br_get_all_2 (ia32_branch **arr, ia32_branch *root, unsigned int *count);

/* ia32_graphviz_br_output_2 - helper function
 */
static void
ia32_graphviz_br_output_2 (FILE *fp, ia32_function *func, ia32_branch *br);


/*** implementation
 */

/*** branch routines
 */

ia32_branch *
ia32_br_new (void)
{
	ia32_branch *	new = calloc (1, sizeof (ia32_branch));

	new->passed = 0;
	new->start = new->end = 0;

	new->last_ilen = new->last_unused = 0;

	new->endtype = BR_END_INVALID;
	new->endbr_count = 0;
	new->endbr = NULL;

	return (new);
}


ia32_branch *
ia32_br_get_unpassed (ia32_branch *root)
{
	ia32_branch **	all;
	ia32_branch *	unpassed = NULL;
	unsigned int	all_count,
			n;


	all = ia32_br_get_all (root, &all_count);
	for (n = 0 ; unpassed == NULL && n < all_count ; ++n)
		if (all[n]->passed == 0)
			unpassed = all[n];
	free (all);

	return (unpassed);
}


ia32_branch *
ia32_br_split (ia32_branch *br, unsigned int vaddr)
{
	ia32_branch *	br_cont = ia32_br_new ();


	/* shallow copy all standard data, then adjust for the split
	 */
	memcpy (br_cont, br, sizeof (ia32_branch));

	br->end = vaddr;
	br_cont->start = vaddr;

	br->endtype = BR_END_PASS;
	br->endbr_count = 1;
	br->endbr = malloc (sizeof (ia32_branch *));
	br->endbr[0] = br_cont;
	printf ("old last_ilen = %d, squashed to zero\n",
		br->last_ilen);
	br->last_ilen = 0;


	return (br_cont);
}


ia32_branch **
ia32_br_get_all (ia32_branch *root, unsigned int *count)
{
	*count = 0;

	return (ia32_br_get_all_2 (NULL, root, count));
}


static ia32_branch **
ia32_br_get_all_2 (ia32_branch **arr, ia32_branch *root, unsigned int *count)
{
	unsigned int	n;


	/* check if cur is already inside the array
	 */
	for (n = 0 ; n < *count ; ++n)
		if (arr[n] == root)
			return (arr);

	/* else add this branch and recurse
	 */
	*count += 1;
	arr = xrealloc (arr, *count * sizeof (ia32_branch *));
	arr[*count - 1] = root;

	for (n = 0 ; n < root->endbr_count ; ++n)
		arr = ia32_br_get_all_2 (arr, root->endbr[n], count);

	return (arr);
}


ia32_branch *
ia32_br_find (ia32_branch *root, unsigned int vaddr)
{
	ia32_branch *	br_hit = NULL;
	ia32_branch **	br_all;
	unsigned int	br_count,
			n;

	br_all = ia32_br_get_all (root, &br_count);

	for (n = 0 ; br_hit == NULL && n < br_count ; ++n) {
		if (br_all[n]->start == vaddr) {
			br_hit = br_all[n];
		} else if (br_all[n]->passed &&
			ia32_trace_range (br_all[n]->start, br_all[n]->end,
				vaddr))
		{
			br_hit = ia32_br_split (br_all[n], vaddr);
		}
	}

	free (br_all);

	return (br_hit);
}


int
ia32_brl_find (ia32_branch **list, unsigned int list_len, ia32_branch *br)
{
	unsigned int	n;


	for (n = 0 ; n < list_len ; ++n)
		if (list[n] == br)
			return (n);

	return (-1);
}


int
ia32_trace_range (unsigned int start, unsigned int end, unsigned int ref)
{
	if (ref < start || ref >= end)
		return (0);

	return (1);
}


unsigned int
ia32_trace_control (ia32_instruction *inst, unsigned char *inst_src,
	unsigned int cur, int *resume, int *ctrl_volatile)
{
	int		resume_dummy,
			vol_dummy;
	unsigned int	std_displ;


	if (resume == NULL)
		resume = &resume_dummy;
	if (ctrl_volatile == NULL)
		ctrl_volatile = &vol_dummy;

	*resume = 0;
	*ctrl_volatile = 0;

	std_displ = inst->length + inst->opc.displ_value + cur;

	if (OD_TEST (inst->opc.used, OP_CONTROL) == 0)
		return (0xffffffff);

	switch (inst_src[0]) {
		/* return */
	case (0xc2):
	case (0xc3):
	case (0xca):
	case (0xcb):
		break;

		/* interrupts, usually syscalls. lets ignore them
		 */
	case (0xcd):
		*resume = 1;
		*ctrl_volatile = 1;
		break;

		/* call with modr/rm, this cannot be predicted statically, so
		 * lets ignore it
		 */
	case (0xff):
		if ((inst_src[1] & 0x38) == 0x10) {
			*resume = 1;
			*ctrl_volatile = 1;
		} else if ((inst_src[1] & 0x38) == 0x20 ||
			(inst_src[1] & 0x38) == 28)
		{
			*resume = 0;
			*ctrl_volatile = 1;
		} else {
			fprintf (stderr, "invalid 0xff control opcode\n");
			exit (EXIT_FAILURE);
		}
		return (0);
		break;

		/* call with displacement */
	case (0xe8):
		*resume = 1;
		return (std_displ);
		break;
		/* jcc with displacement */
	case (0x70):
	case (0x71):
	case (0x72):
	case (0x73):
	case (0x74):
	case (0x75):
	case (0x76):
	case (0x77):
	case (0x78):
	case (0x79):
	case (0x7a):
	case (0x7b):
	case (0x7c):
	case (0x7d):
	case (0x7e):
	case (0x7f):
		return (std_displ);
		break;
		/* jcc with full displacement */
	case (0x0f):
		if ((inst_src[1] & 0xf0) == 0x80)
			return (std_displ);
		break;
		/* jcxz */
	case (0xe3):
		return (std_displ);
		break;
		/* jump short and direct */
	case (0xeb):
	case (0xe9):
		return (std_displ);
		break;
		/* jmp direct intersegment */
	case (0xea):
		return (inst->opc.imm_value);
		break;
		/* loop* */
	case (0xe0):
	case (0xe1):
	case (0xe2):
		return (std_displ);
		break;
	default:
		fprintf (stderr, "invalid control flow instruction\n");
		exit (EXIT_FAILURE);
		break;
	}

	return (0xffffffff);
}


void
ia32_graphviz_br_output (FILE *fp, ia32_branch *root, void *func_2)
{
	ia32_function *	func = (ia32_function *) func_2;
	ia32_branch **	br_all;
	unsigned int	br_count,
			n,
			i;
	ia32_branch *	brw;	/* working branch */


	fprintf (fp, "digraph %s {\n",
		(func->name != NULL) ? func->name : "__unknown");
#ifdef	DIN_A4
	fprintf (fp, "\tsize = \"8.2,11.0\"\n");
#endif

	fprintf (fp, "\tnode [\n");
	fprintf (fp, "\t\tstyle = filled\n");
	fprintf (fp, "\t\tshape = \"record\"\n");
	fprintf (fp, "\t\tfillcolor = \"lightskyblue\"\n");
	fprintf (fp, "\t];\n");

	br_all = ia32_br_get_all (root, &br_count);
	for (n = 0 ; n < br_count ; ++n) {
		brw = br_all[n];

		fprintf (fp, "\t\"0x%08x\" [\n", brw->start);
		fprintf (fp, "\t\tlabel = \"{ <fi> 0x%08x | { ",
			brw->start);

		ia32_graphviz_br_output_2 (fp, func, brw);

		fprintf (fp, "} | ");

#ifdef	XREF_PRINT
		for (i = 0 ; i < func->func_xref_count ; ++i)
			fprintf (fp, "xref: 0x%08x to 0x%08x\\l",
				func->func_xref[i]->from,
				func->func_xref[i]->to);
		fprintf (fp, "| ");
#endif

		fprintf (fp, "<fo> 0x%08x }\"\n", brw->end);

		/* find branches that contain external function references,
		 * and color those differently. color call branches red before.
		 */
		if (brw->endtype == BR_END_CALL) {
			fprintf (fp, "\t\tfillcolor = \"red\"\n");
		} else if (brw->endtype == BR_END_UNPREDICT) {
			fprintf (fp, "\t\tfillcolor = \"orange\"\n");
		} else for (i = 0 ; i < func->func_xref_count ; ++i) {
			if (ia32_trace_range (brw->start, brw->end,
				func->func_xref[i]->from))
			{
				if (func->func_xref[i]->to == func->start)
					fprintf (fp, "\t\tfillcolor = \"limegreen\"\n");
				else
					fprintf (fp, "\t\tfillcolor = \"mediumturquoise\"\n");
			}
		}

		fprintf (fp, "\t];\n");
	}

	for (n = 0 ; n < br_count ; ++n) {
		brw = br_all[n];

		for (i = 0 ; i < brw->endbr_count ; ++i) {
			fprintf (fp, "\t\"0x%08x\":fo -> \"0x%08x\":fi",
				brw->start, brw->endbr[i]->start);

			if (brw->endtype == BR_END_PASS)
				fprintf (fp, " [color=\"seagreen\"]");
			else if (brw->endtype == BR_END_IF)
				fprintf (fp, " [color=\"%s\"]",
					i == 0 ? "seagreen" : "maroon3");

			fprintf (fp, ";\n");
		}
	}

	fprintf (fp, "}\n");

	free (br_all);

	return;
}


static void
ia32_graphviz_br_output_2 (FILE *fp, ia32_function *func, ia32_branch *br)
{
	ia32_instruction *	inst,	/* helper pointer for inst_s */
				inst_s;	/* instruction structure */
	unsigned int		vaddr;	/* current virtual address */
	unsigned char *		cur;	/* memory position of vaddr */
	unsigned int		ilen = 0;
	char			inst_str[128];


	for (vaddr = br->start ; vaddr < br->end ; vaddr += ilen) {
		cur = &func->mem[vaddr - func->start];
		inst = ia32_decode_instruction (cur, &inst_s);

		if (inst == NULL)
			break;
		
		fprintf (fp, "\xff""0x%08x\\l", vaddr);
		ilen = inst->length;
	}

	fprintf (fp, "|\xff");
	for (vaddr = br->start ; vaddr < br->end ; vaddr += ilen) {
		cur = &func->mem[vaddr - func->start];
		inst = ia32_decode_instruction (cur, &inst_s);
		
		assert (inst != NULL);

		ia32_sprint (inst, inst_str, sizeof (inst_str));
		fprintf (fp, "%s\\l ", inst_str);
		ilen = inst->length;
	}

	return;
}



#ifdef	TESTING

int
main (int argc, char *argv[])
{
	char *		fn;
	unsigned int	amount;
	FILE *		fp;
	unsigned char	buf[4096];
	ia32_function *	func;


	printf ("ia32 control flow tracer "IA32_TRACE_VERSION"\n");

	if (argc < 2) {
		fprintf (stderr, "usage: %s file.bin [file.dot]\n\n", argv[0]);

		exit (EXIT_FAILURE);
	}

	fp = fopen (argv[1], "rb");
	if (fp == NULL) {
		perror ("fopen");

		exit (EXIT_FAILURE);
	}


	for (fn = argv[1] ; *fn != '\0' && *fn != '.' ; ++fn)
		;
	if (*fn == '.')
		*fn = '\0';

	amount = fread (buf, 1, sizeof (buf), fp);
	fclose (fp);

	func = ia32_func_new ();
	func->name = argv[1];
	func->mem = buf;
	func->start = 0x01000000;
	func->end = func->start + amount;
	func->br_root = NULL;
	func->func_xref_count = 0;
	func->func_xref = NULL;

	func->br_root = ia32_func_breakup (func, NULL);

	fp = fopen (argc >= 3 ? argv[2] : "test.dot", "w");
	if (fp == NULL) {
		perror ("fopen");

		exit (EXIT_FAILURE);
	}

	ia32_graphviz_br_output (fp, func->br_root, func);
	fclose (fp);

	exit (EXIT_SUCCESS);
}

#endif


