/* ia32-trace.h - ia32 control flow tracer include file
 *
 * by scut
 */

#ifndef	IA32_TRACE_H
#define	IA32_TRACE_H

#define	IA32_TRACE_VERSION	"0.2.2"

#include <ia32-decode.h>


/* XXX: clash */
/* #include <elf-glue.h>*/


/* ia32_branch
 *
 * one 'branch', which we use in the sense of "one continous block of code"
 * here. a branch is always executed from top to bottom, without any
 * conditions. at the end of a branch a new branch is choosen to continue
 * with, or the branch terminates. what happens is defined by `endtype',
 * where we possibly go by `endrb', which is `endbr_count' items long.
 *
 * functions that operate on branches are ia32_br_*
 */
typedef struct ia32_branch {
	int		passed;		/* 0 (not processed yet), 1 (done) */
	unsigned int	start;		/* virt. start address */
	unsigned int	end;		/* virt. end address (if passed = 1) */

	unsigned int	last_ilen;	/* length of last instruction */
	int		last_unused;	/* if extended, the free space @ end */

	void *		user;		/* user data pointer, free to use */

	/* _END_INVALID =
	 * the control flow tracer could not successfully continue, most likely
	 * an instruction could not be decoded
	 */
#define	BR_END_INVALID	0
	/* _END_RET = INTER, UNCOND, PASS
	 * a normal return from the function
	 */
#define	BR_END_RET	1
	/* _END_IF = INTRA, COND, PASS
	 * an if statement within the function
	 * also the `cond' field is used when _END_IF is endtype.
	 * when the condition evaluates to true then the second branch is
	 * choosen (endbr[1]), else the first.
	 */
#define	BR_END_IF	2
	/* _END_PASS =
	 * there is only one end branch, and we unconditionally go with it.
	 * this happens when a branch is split due to another branch
	 * referencing into it. it can also happen for control flow
	 * instructions we cannot predict or that we intentionally ignore,
	 * such as: call r/m, int, ..
	 */
#define	BR_END_PASS	3
	/* _END_TRANSFER =
	 * similar to _END_PASS, this unconditionally transfers control to
	 * another position within the function. the difference to _END_PASS
	 * is that it does change the control flow, while _END_PASS continues
	 * at the instruction right after the branch.
	 */
#define	BR_END_TRANSFER	4
	/* _END_CALL
	 * similar to _END_PASS, this resumes execution. the difference is
	 * that it always resumes directly after the current instruction,
	 * which is call to another function.
	 */
#define	BR_END_CALL	5
	/* _END_IF_INTER
	 * like if, but a conditional jump refering to outside of the function
	 * itself. rarely used, but libc uses it for example to cope faster
	 * with error conditions (__syscall_error)
	 */
#define	BR_END_IF_INTER	6
	/* _END_UNPREDICT
	 * is used when we hit a control flow instruction that we cannot
	 * predict statically and that does not return. this rather marks
	 * a branch end more than it affects any control flow (we just run the
	 * instruction and see what happens).
	 */
#define	BR_END_UNPREDICT	7

	int		endtype;

	unsigned int	cond;

	unsigned int	other_xref_count;
	void *		other_xref;	/* type: ia32_xref ** */

	/* endbr[] is an array of branches possibly taken at the end of this
	 * branch. there are endbr_count branches (which should not exceed 2)
	 *
	 * for endbr_count = 0, endbr is guaranteed to be NULL.
	 */
	unsigned int		endbr_count;
	struct ia32_branch **	endbr;
} ia32_branch;


/*** function prototypes
 */

/* ia32_br_new
 *
 * create a new branch structure.
 *
 * return pointer to new structure
 */

ia32_branch *
ia32_br_new (void);


/* ia32_br_get_unpassed
 *
 * recursivly check all branches reachable from `root' whether they have been
 * processed already.
 *
 * return the first unpassed branch found on success
 * return NULL when all branches have been processed
 */

ia32_branch *
ia32_br_get_unpassed (ia32_branch *root);


/* ia32_br_split
 *
 * split the branch `br' at `vaddr', which has to be inbetween br->start and
 * br->end. hence the branch should be processed already, else the result is
 * undefined.
 *
 * return the second branch, which was just created due to the splitting
 */

ia32_branch *
ia32_br_split (ia32_branch *br, unsigned int vaddr);


/* ia32_br_get_all
 *
 * walk all branches recursivly from `root', collecting pointers to all
 * branches. store the number of unique branches found into `br_count'.
 *
 * return array of pointers to branches, which is `br_count' entries long
 */

ia32_branch **
ia32_br_get_all (ia32_branch *root, unsigned int *count);


/* ia32_br_find
 *
 * find the correct branch which starts at `vaddr', traveling the parse tree
 * down from the root branch `root'. if a branch is found that covers the
 * address, but does not start there, the branch is automatically split.
 *
 * return pointer to branch starting at `vaddr' on success
 * return NULL if no branch has been found (you have to create a new one) on
 *     failure
 */

ia32_branch *
ia32_br_find (ia32_branch *root, unsigned int vaddr);


/* ia32_brl_find
 *
 * utility function, locate `br' in branch list `list', which is `list_len'
 * items long.
 *
 * return index into `list' on success
 * return -1 on failure
 */

int
ia32_brl_find (ia32_branch **list, unsigned int list_len, ia32_branch *br);


/* ia32_trace_range
 *
 * test whether `ref' lies inbetween `start' and `end'.
 *
 * return 0 if it is outside of [start - end[
 * return != 0 if it is within
 */

int
ia32_trace_range (unsigned int start, unsigned int end, unsigned int ref);


/* ia32_trace_control
 *
 * decode the control flow instruction `inst' which was found at virtual
 * address `cur'. the instruction was found in memory at `inst_src'.
 * try to tell the destination of the control instruction as a virtual address
 * control will be transfered. for conditional instructions, the condition is
 * always assumed true. `resume' contains 0 or 1 after the function returns,
 * where 1 means the control flow will resume after the instruction at some
 * point (i.e. call), where 0 denotes that no information about control
 * transfer origin is kept. `resume' can be NULL, then this information is
 * not passed to the caller. `ctrl_volatile' tags runtime-dependant
 * instructions, such as "call eax" and "call [0x0804826c]", etc.
 *
 * return absolute virtual address control may be passed to on success
 * return 0xffffffff on failure
 */

unsigned int
ia32_trace_control (ia32_instruction *inst, unsigned char *inst_src,
	unsigned int cur, int *resume, int *ctrl_volatile);


#include <ia32-function.h>

/* ia32_graphviz_br_output
 *
 * output a branch tree starting at `root' in graphviz format to the file
 * mapped by `fp', which has to be open and will remain open. the branch is
 * within `func'. when `digraph' is not zero, a digraph is created, else
 * a subgraph.
 *
 * return in any case
 */
#if 0
void
ia32_graphviz_br_output (FILE *fp, ia32_branch *root, ia32_function *func);
#endif

/* FIXME: find a cleaner way to work around header file deps
 */
void
ia32_graphviz_br_output (FILE *fp, ia32_branch *root, void *func);

#endif


